@@ -279,7 +279,6 @@ grn_dat_create(grn_ctx *ctx, const char *path, uint32_t,
     return NULL;
   }
   grn_dat_init(ctx, dat);
-  dat->obj.header.flags = flags;
 
   dat->io = grn_io_create(ctx, path, sizeof(struct grn_dat_header),
                           4096, 0, grn_io_auto, GRN_IO_EXPIRE_SEGMENT);
@@ -302,8 +301,19 @@ grn_dat_create(grn_ctx *ctx, const char *path, uint32_t,
   dat->header->encoding = encoding;
   dat->header->tokenizer = GRN_ID_NIL;
   dat->header->file_id = 0;
+  if (dat->header->flags & GRN_OBJ_KEY_NORMALIZE) {
+    dat->header->flags &= ~GRN_OBJ_KEY_NORMALIZE;
+    dat->header->normalizer = GRN_DB_NORMALIZER_AUTO;
+    dat->normalizer = grn_ctx_at(ctx, dat->header->normalizer);
+  } else {
+    dat->header->normalizer = GRN_ID_NIL;
+    dat->normalizer = NULL;
+  }
   dat->encoding = encoding;
   dat->tokenizer = NULL;
+
+  dat->obj.header.flags = dat->header->flags;
+
   return dat;
 }
 
@@ -337,6 +347,11 @@ grn_dat_open(grn_ctx *ctx, const char *path)
   dat->encoding = dat->header->encoding;
   dat->obj.header.flags = dat->header->flags;
   dat->tokenizer = grn_ctx_at(ctx, dat->header->tokenizer);
+  if (dat->header->flags & GRN_OBJ_KEY_NORMALIZE) {
+    dat->header->flags &= ~GRN_OBJ_KEY_NORMALIZE;
+    dat->header->normalizer = GRN_DB_NORMALIZER_AUTO;
+  }
+  dat->normalizer = grn_ctx_at(ctx, dat->header->normalizer);
   return dat;
 }
 
@@ -648,9 +663,11 @@ grn_dat_update(grn_ctx *ctx, grn_dat *dat,
   return GRN_SUCCESS;
 }
 
-int grn_dat_scan(grn_ctx *ctx, grn_dat *dat, const char *str,
-                         unsigned int str_size, grn_dat_scan_hit *scan_hits,
-                         unsigned int max_num_scan_hits, const char **str_rest) {
+int
+grn_dat_scan(grn_ctx *ctx, grn_dat *dat, const char *str,
+             unsigned int str_size, grn_dat_scan_hit *scan_hits,
+             unsigned int max_num_scan_hits, const char **str_rest)
+{
   if (!grn_dat_open_trie_if_needed(ctx, dat) || !str ||
       !(dat->obj.header.flags & GRN_OBJ_KEY_VAR_SIZE) || !scan_hits) {
     return -1;
@@ -670,16 +687,17 @@ int grn_dat_scan(grn_ctx *ctx, grn_dat *dat, const char *str,
 
   int num_scan_hits = 0;
   try {
-    if (dat->obj.header.flags & GRN_OBJ_KEY_NORMALIZE) {
-      grn_str * const normalized_str = grn_str_open(
-          ctx, str, str_size, GRN_STR_NORMALIZE | GRN_STR_WITH_CHECKS);
-      if (!normalized_str) {
-        fprintf(stderr, "error: grn_str_open() failed!\n");
+    if (dat->normalizer) {
+      int flags = GRN_STRING_WITH_CHECKS;
+      grn_obj * const normalized_string = grn_string_open(ctx, str, str_size,
+                                                          dat->normalizer,
+                                                          flags);
+      if (!normalized_string) {
+        fprintf(stderr, "error: grn_string_open() failed!\n");
         return -1;
       }
-      str = normalized_str->norm;
-      str_size = normalized_str->norm_blen;
-      const short *checks = normalized_str->checks;
+      grn_string_get_normalized(ctx, normalized_string, &str, &str_size, NULL);
+      const short *checks = grn_string_get_checks(ctx, normalized_string);
       unsigned int offset = 0;
       while (str_size) {
         if (*checks) {
@@ -716,9 +734,10 @@ int grn_dat_scan(grn_ctx *ctx, grn_dat *dat, const char *str,
         ++checks;
       }
       if (str_rest) {
-        *str_rest = normalized_str->orig + offset;
+        grn_string_get_original(ctx, normalized_string, str_rest, NULL);
+        *str_rest += offset;
       }
-      grn_str_close(ctx, normalized_str);
+      grn_obj_close(ctx, normalized_string);
     } else {
       const char * const begin = str;
       while (str_size) {
