@@ -257,7 +257,7 @@ grn_tiny_bitmap_put_and_set(grn_tiny_bitmap *bitmap, grn_id bit_id,
 
 /* grn_io_array */
 
-inline void *
+inline static void *
 grn_io_array_at_inline(grn_ctx *ctx, grn_io *io, uint32_t segment_id,
                        uint32_t offset, int flags)
 {
@@ -318,9 +318,26 @@ grn_io_array_bit_flip(grn_ctx *ctx, grn_io *io,
   return ptr;
 }
 
+/* grn_table_queue */
+
+static void
+grn_table_queue_lock_init(grn_ctx *ctx, grn_table_queue *queue)
+{
+  MUTEX_INIT_SHARED(queue->mutex);
+  COND_INIT_SHARED(queue->cond);
+}
+
+static void
+grn_table_queue_init(grn_ctx *ctx, grn_table_queue *queue)
+{
+  queue->head = 0;
+  queue->tail = 0;
+  queue->cap = GRN_ID_MAX;
+  grn_table_queue_lock_init(ctx, queue);
+}
+
 /* grn_array */
 
-#define GRN_ARRAY_HEADER_SIZE  0x9000
 #define GRN_ARRAY_SEGMENT_SIZE 0x400000
 
 /* Header of grn_io-based grn_array. */
@@ -332,7 +349,8 @@ struct grn_array_header {
   uint32_t n_garbages;
   grn_id garbages;
   uint32_t lock;
-  uint32_t reserved[5];
+  uint32_t reserved[9];
+  grn_table_queue queue;
 };
 
 /*
@@ -443,7 +461,7 @@ grn_array_init_io_array(grn_ctx *ctx, grn_array *array, const char *path,
   header->n_entries = 0;
   header->n_garbages = 0;
   header->garbages = GRN_ID_NIL;
-
+  grn_table_queue_init(ctx, &header->queue);
   array->obj.header.flags = flags;
   array->ctx = ctx;
   array->value_size = value_size;
@@ -457,6 +475,26 @@ grn_array_init_io_array(grn_ctx *ctx, grn_array *array, const char *path,
   return GRN_SUCCESS;
 }
 
+void
+grn_array_queue_lock_clear(grn_ctx *ctx, grn_array *array)
+{
+  struct grn_array_header *header;
+  header = grn_io_header(array->io);
+  grn_table_queue_lock_init(ctx, &header->queue);
+}
+
+grn_table_queue *
+grn_array_queue(grn_ctx *ctx, grn_array *array)
+{
+  if (grn_array_is_io_array(array)) {
+    struct grn_array_header *header;
+    header = grn_io_header(array->io);
+    return &header->queue;
+  } else {
+    return NULL;
+  }
+}
+
 static grn_rc
 grn_array_init(grn_ctx *ctx, grn_array *array,
                const char *path, uint32_t value_size, uint32_t flags)
@@ -642,7 +680,7 @@ grn_array_set_value_inline(grn_ctx *ctx, grn_array *array, grn_id id,
     return GRN_SUCCESS;
   case GRN_OBJ_INCR :
     switch (array->value_size) {
-  case sizeof(int32_t) :
+    case sizeof(int32_t) :
       *((int32_t *)entry) += *((int32_t *)value);
       return GRN_SUCCESS;
     case sizeof(int64_t) :
@@ -987,6 +1025,13 @@ grn_array_add_to_io_array(grn_ctx *ctx, grn_array *array, void **value)
   return id;
 }
 
+void
+grn_array_clear_curr_rec(grn_ctx *ctx, grn_array *array)
+{
+  struct grn_array_header * const header = array->header;
+  header->curr_rec = GRN_ID_NIL;
+}
+
 grn_id
 grn_array_add(grn_ctx *ctx, grn_array *array, void **value)
 {
@@ -1437,8 +1482,17 @@ grn_io_hash_init(grn_ctx *ctx, grn_hash *hash, const char *path,
   header->n_entries = 0;
   header->n_garbages = 0;
   header->tokenizer = GRN_ID_NIL;
+  if (header->flags & GRN_OBJ_KEY_NORMALIZE) {
+    header->flags &= ~GRN_OBJ_KEY_NORMALIZE;
+    header->normalizer = GRN_DB_NORMALIZER_AUTO;
+    hash->normalizer = grn_ctx_at(ctx, header->normalizer);
+  } else {
+    header->normalizer = GRN_ID_NIL;
+    hash->normalizer = NULL;
+  }
+  grn_table_queue_init(ctx, &header->queue);
 
-  hash->obj.header.flags = flags;
+  hash->obj.header.flags = header->flags;
   hash->ctx = ctx;
   hash->key_size = key_size;
   hash->encoding = encoding;
@@ -1509,6 +1563,7 @@ grn_tiny_hash_init(grn_ctx *ctx, grn_hash *hash, const char *path,
   hash->n_entries_ = 0;
   hash->garbages = GRN_ID_NIL;
   hash->tokenizer = NULL;
+  hash->normalizer = NULL;
   grn_tiny_array_init(ctx, &hash->a, entry_size, GRN_TINY_ARRAY_CLEAR);
   grn_tiny_bitmap_init(ctx, &hash->bitmap);
   return GRN_SUCCESS;
@@ -1575,6 +1630,11 @@ grn_hash_open(grn_ctx *ctx, const char *path)
             hash->header = header;
             hash->lock = &header->lock;
             hash->tokenizer = grn_ctx_at(ctx, header->tokenizer);
+            if (header->flags & GRN_OBJ_KEY_NORMALIZE) {
+              header->flags &= ~GRN_OBJ_KEY_NORMALIZE;
+              header->normalizer = GRN_DB_NORMALIZER_AUTO;
+            }
+            hash->normalizer = grn_ctx_at(ctx, header->normalizer);
             return hash;
           } else {
             GRN_LOG(ctx, GRN_LOG_NOTICE,
@@ -1821,7 +1881,7 @@ grn_hash_lock(grn_ctx *ctx, grn_hash *hash, int timeout)
           GRN_LOG(ctx, GRN_LOG_NOTICE, "hash(%p) collisions(%d/%d)", hash, _ncolls, _ncalls);
         }
       }
-      usleep(1000);
+      grn_nanosleep(1000000);
       continue;
     }
     return GRN_SUCCESS;
@@ -2876,7 +2936,7 @@ grn_hash_check(grn_ctx *ctx, grn_hash *hash)
   char buf[8];
   struct grn_hash_header *h = hash->header;
   GRN_OUTPUT_ARRAY_OPEN("RESULT", 1);
-  GRN_OUTPUT_MAP_OPEN("SUMMARY", 24);
+  GRN_OUTPUT_MAP_OPEN("SUMMARY", 25);
   GRN_OUTPUT_CSTR("flags");
   grn_itoh(h->flags, buf, 8);
   GRN_OUTPUT_STR(buf, 8);
@@ -2886,6 +2946,8 @@ grn_hash_check(grn_ctx *ctx, grn_hash *hash)
   GRN_OUTPUT_INT64(hash->value_size);
   GRN_OUTPUT_CSTR("tokenizer");
   GRN_OUTPUT_INT64(h->tokenizer);
+  GRN_OUTPUT_CSTR("normalizer");
+  GRN_OUTPUT_INT64(h->normalizer);
   GRN_OUTPUT_CSTR("curr_rec");
   GRN_OUTPUT_INT64(h->curr_rec);
   GRN_OUTPUT_CSTR("curr_key");
